@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2007 NTT DATA CORPORATION.
+ * Copyright (c) 2006-2008 NTT DATA CORPORATION.
  * All rights reserved.
  */
 
@@ -16,6 +16,7 @@
 PG_FUNCTION_INFO_V1(pgs2pdftotext1);
 PG_FUNCTION_INFO_V1(pgs2pdftotext2);
 PG_FUNCTION_INFO_V1(pgs2snippet1);
+PG_FUNCTION_INFO_V1(pgs2norm);
 PG_FUNCTION_INFO_V1(pgs2textporter1);
 PG_FUNCTION_INFO_V1(pgs2textporter2);
 
@@ -24,7 +25,7 @@ inline static void
 tempfile_unlink(char *path)
 {
   if (unlink(path) == -1) {
-    elog(ERROR, "pgsenna2: failed to unlink temporary file (%d)", errno);
+    elog(ERROR, "pgsenna2: failed to unlink temporary file (%s)", strerror(errno));
   }
 }
 
@@ -45,12 +46,9 @@ pgs2pdftotext1(PG_FUNCTION_ARGS)
   command[511] = '\0';
   stdout_pdftotext = popen(command, "r");
   if (stdout_pdftotext == (FILE *)-1) {
-    elog(ERROR, "pgsenna2: failed to popen for pdftotext (%d)", errno);
-  }
-  filtered = palloc(VARHDRSZ + buf_size);
-  if (filtered == NULL) {
-    elog(ERROR, "pgsenna2: failed to palloc for return value");
+    elog(ERROR, "pgsenna2: failed to popen for pdftotext (%s)", strerror(errno));
   }
+  filtered = PGS2_PALLOC(VARHDRSZ + buf_size);
   while ((buf_read = fread(VARDATA(filtered) + buf_read_total,
                            sizeof(char), buf_size - buf_read_total,
                            stdout_pdftotext))) {
@@ -67,7 +65,7 @@ pgs2pdftotext1(PG_FUNCTION_ARGS)
     elog(ERROR, "pgsenna2: failed to fread temporary PDFfile");
   }
   if (pclose(stdout_pdftotext) != 0) {
-    elog(ERROR, "pgsenna2: failed pdftotext 1 (%d)", errno);
+    elog(ERROR, "pgsenna2: failed pdftotext 1 (%s)", strerror(errno));
   }
   /* varatt_size include the size of itself */
 #ifdef POSTGRES83
@@ -98,13 +96,13 @@ pgs2pdftotext2(PG_FUNCTION_ARGS)
   fd = mkstemp(path);
 #endif
   if (fd == -1) {
-    elog(ERROR, "pgsenna2: failed mkstemp for temporary PDFfile (%d)", errno);
+    elog(ERROR, "pgsenna2: failed mkstemp for temporary PDFfile (%s)", strerror(errno));
   }
   close(fd);
   tempfile = fopen(path, "wb");
   if (tempfile == NULL) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed to fopen temporary PDFfile (%d)", errno);
+    elog(ERROR, "pgsenna2: failed to fopen temporary PDFfile (%s)", strerror(errno));
   }
   while ((byte_wrote = fwrite(VARDATA(pdfdata) + byte_wrote_total, sizeof(char),
                               VARSIZE(pdfdata) - VARHDRSZ - byte_wrote_total,
@@ -113,7 +111,7 @@ pgs2pdftotext2(PG_FUNCTION_ARGS)
   }
   if (fclose(tempfile) != 0) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed to fclose temporary PDFfile (%d)", errno);
+    elog(ERROR, "pgsenna2: failed to fclose temporary PDFfile (%s)", strerror(errno));
   }
   snprintf(command, sizeof(command), "pdftotext %s -", path);
   //  snprintf(command, sizeof(command), "wvWare --charset UTF-8 %s -", path);
@@ -122,7 +120,7 @@ pgs2pdftotext2(PG_FUNCTION_ARGS)
   stdout_pdftotext = popen(command, "r");
   if (stdout_pdftotext == (FILE *)-1) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed to popen for pdftotext (%d)", errno);
+    elog(ERROR, "pgsenna2: failed to popen for pdftotext (%s)", strerror(errno));
   }
   filtered = palloc(VARHDRSZ + buf_size);
   if (filtered == NULL) {
@@ -148,7 +146,7 @@ pgs2pdftotext2(PG_FUNCTION_ARGS)
   }
   if (pclose(stdout_pdftotext) != 0) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed pdftotext 2 (%d)", errno);
+    elog(ERROR, "pgsenna2: failed pdftotext 2 (%s)", strerror(errno));
   }
   /* varatt_size include the size of itself */
 #ifdef POSTGRES83
@@ -167,9 +165,9 @@ Datum pgs2snippet1(PG_FUNCTION_ARGS)
   size_t width = (size_t)PG_GETARG_INT32(1);
   unsigned int max_results = PG_GETARG_INT32(2);
   text *defaultopentag_ = (text*)PG_GETARG_TEXT_P(3);
-  char *defaultopentag = text2cstr(defaultopentag_);
+  const char *defaultopentag = text2cstr(defaultopentag_);
   text *defaultclosetag_ = (text*)PG_GETARG_TEXT_P(4);
-  char *defaultclosetag = text2cstr(defaultclosetag_);
+  const char *defaultclosetag = text2cstr(defaultclosetag_);
   sen_snip_mapping *mapping = (sen_snip_mapping *)PG_GETARG_POINTER(5);
   text *keywords_ = (text*)PG_GETARG_TEXT_P(6);
   char *keywords = text2cstr(keywords_);
@@ -180,56 +178,46 @@ Datum pgs2snippet1(PG_FUNCTION_ARGS)
   sen_snip *snip = NULL;
   unsigned int nresults = 0;
   unsigned int max_tagged_len = 0;
-  char *tokenp = NULL;
-  sen_encoding encoding = sen_enc_default;
+  unsigned int strlen_dot = strlen(defaultopentag);
+  unsigned int strlen_dct = strlen(defaultclosetag);
 
-  switch (GetDatabaseEncoding()) {
-  case PG_UTF8:
-    encoding = sen_enc_utf8;
-    break;
-  case PG_EUC_JP:
-    encoding = sen_enc_euc_jp;
-    break;
-  case PG_SJIS:
-    encoding = sen_enc_sjis;
-    break;
-  default:
-    encoding = sen_enc_default;
-  }
-  snip = sen_snip_open(encoding, flags, width, max_results,
-                       defaultopentag, strlen(defaultopentag),
-                       defaultclosetag, strlen(defaultclosetag),
-                       mapping);
+  snip = sen_query_snip(get_query(keywords), flags, width, max_results, 1,
+                        &defaultopentag, &strlen_dot,
+                        &defaultclosetag, &strlen_dct,
+                        mapping);
   if (!snip) {
-    elog(ERROR, "pgsenna2: sen_snip_open() failed");
-  }
-  tokenp = strtok(keywords, " ");
-  while (tokenp != NULL) {
-    rc = sen_snip_add_cond(snip, tokenp, strlen(tokenp), NULL, 0, NULL, 0);
-    if (rc != sen_success) {
-      elog(ERROR, "pgsenna2: sen_snip_add_cond() failed %d", rc);
-    }
-    tokenp = strtok(NULL, " ");
+    elog(ERROR, "pgsenna2: sen_query_snip() failed");
   }
   rc = sen_snip_exec(snip, document, strlen(document),
                      &nresults, &max_tagged_len);
   if (rc != sen_success) {
     elog(ERROR, "pgsenna2: sen_snip_exec() failed %d", rc);
   }
-  result = palloc(VARHDRSZ + max_tagged_len);
+  result = PGS2_PALLOC(VARHDRSZ + max_tagged_len);
   memset(VARDATA(result), 0, max_tagged_len);
   rc = sen_snip_get_result(snip, 0, VARDATA(result), &result_len);
-  if (rc != sen_success) {
+  if (rc == sen_invalid_argument) {
+    rc = sen_snip_close(snip);
+    if (rc != sen_success) {
+      elog(ERROR, "pgsenna2: sen_snip_close() failed1 %d", rc);
+    }
+    pfree((char *)defaultopentag);
+    pfree((char *)defaultclosetag);
+    pfree(keywords);
+    pfree(document);
+    PG_RETURN_NULL();
+  } else if (rc != sen_success) {
     elog(ERROR, "pgsenna2: sen_snip_get_result() failed %d", rc);
   }
   rc = sen_snip_close(snip);
   if (rc != sen_success) {
-    elog(ERROR, "pgsenna2: sen_snip_close() failed %d", rc);
+    elog(ERROR, "pgsenna2: sen_snip_close() failed2 %d", rc);
   }
-  pfree(defaultopentag);
-  pfree(defaultclosetag);
+  pfree((char *)defaultopentag);
+  pfree((char *)defaultclosetag);
   pfree(keywords);
   pfree(document);
+  max_tagged_len --; /* cutting null-end */
 #ifdef POSTGRES83
   SET_VARSIZE_4B(result, VARHDRSZ + max_tagged_len);
 #else
@@ -238,6 +226,33 @@ Datum pgs2snippet1(PG_FUNCTION_ARGS)
   PG_RETURN_TEXT_P(result);
 }
 
+Datum pgs2norm(PG_FUNCTION_ARGS)
+{
+  const char *source = text2cstr((text*)PG_GETARG_TEXT_P(0));
+  int source_len;
+  int flags = PG_GETARG_INT32(1);
+  text *filtered;
+  int filtered_len;
+
+  if (flags != 0 && flags != 1) {
+    elog(ERROR, "pgsenna2: invalid value at pgs2norm");
+  }
+  source_len = strlen(source);
+  filtered = (text *) PGS2_PALLOC(VARHDRSZ + source_len);
+  filtered_len = sen_str_normalize(source, source_len, get_sen_encoding(),
+                                   flags, VARDATA(filtered), source_len);
+  if (filtered_len == -1) {
+    elog(ERROR, "pgsenna2: sen_nstr_open failed during pgs2norm");
+  }
+  pfree((char *)source);
+#ifdef POSTGRES83
+  SET_VARSIZE_4B(filtered, VARHDRSZ + filtered_len);
+#else
+  VARATT_SIZEP(filtered) = VARHDRSZ + filtered_len;
+#endif
+  PG_RETURN_TEXT_P(filtered);
+}
+
 #ifdef TEXTPORTER
 int _DMC_GetText_V4(unsigned char *pdfpath, unsigned char *path2,
                     char *enc, int flag);
@@ -334,7 +349,7 @@ pgs2textporter1(PG_FUNCTION_ARGS)
   fd = mkstemp(path2);
 #endif
   if (fd == -1) {
-    elog(ERROR, "pgsenna2: failed mkstemp for temporary file 2(%d)", errno);
+    elog(ERROR, "pgsenna2: failed mkstemp for temporary file 2(%s)", strerror(errno));
   }
   close(fd);
   {
@@ -410,13 +425,13 @@ pgs2textporter2(PG_FUNCTION_ARGS)
   fd = mkstemp(path);
 #endif
   if (fd == -1) {
-    elog(ERROR, "pgsenna2: failed mkstemp for temporary file (%d)", errno);
+    elog(ERROR, "pgsenna2: failed mkstemp for temporary file (%s)", strerror(errno));
   }
   close(fd);
   tempfile = fopen(path, "wb");
   if (tempfile == NULL) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed to fopen temporary file (%d)", errno);
+    elog(ERROR, "pgsenna2: failed to fopen temporary file (%s)", strerror(errno));
   }
   while ((byte_wrote = fwrite(VARDATA(pdfdata) + byte_wrote_total, sizeof(char),
                               VARSIZE(pdfdata) - VARHDRSZ - byte_wrote_total,
@@ -425,14 +440,14 @@ pgs2textporter2(PG_FUNCTION_ARGS)
   }
   if (fclose(tempfile) != 0) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed to fclose temporary file (%d)", errno);
+    elog(ERROR, "pgsenna2: failed to fclose temporary file (%s)", strerror(errno));
   }
 #ifndef WIN32
   fd = mkstemp(path2);
 #endif
   if (fd == -1) {
     tempfile_unlink(path);
-    elog(ERROR, "pgsenna2: failed mkstemp for temporary file 2(%d)", errno);
+    elog(ERROR, "pgsenna2: failed mkstemp for temporary file 2(%s)", strerror(errno));
   }
   close(fd);
   {
