@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2007 NTT DATA CORPORATION.
+ * Copyright (c) 2006-2008 NTT DATA CORPORATION.
  * All rights reserved.
  *
  * derived from original pg_senna.c developed by Daisuke Maki
@@ -66,30 +66,35 @@
 #include <dirent.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include <mb/pg_wchar.h>
-#include "access/genam.h"
 #include "access/heapam.h"
-#include "pgsenna2.h"
+#include "access/genam.h"
 #include "access/xlogutils.h"
-#include "catalog/index.h"
 #include "catalog/catalog.h"
-#include "storage/smgr.h"
-#include "storage/lock.h"
+#include "catalog/index.h"
 #include "commands/vacuum.h"
-#include "executor/executor.h"
+#include "funcapi.h"
 #include "miscadmin.h"
+#include "nodes/relation.h"
 #include "optimizer/cost.h"
 #include "optimizer/restrictinfo.h"
-#include "nodes/relation.h"
+#include "pgstat.h"
 #include "storage/itemptr.h"
+#include "storage/lock.h"
+#include "storage/smgr.h"
 #include "utils/guc.h"
-#include "funcapi.h"
 #include "utils/relcache.h"
+#include "utils/selfuncs.h"
+#include "pgsenna2.h"
 
-#if defined(POSTGRES81) && defined(WIN32)
+#ifdef WIN32
+#ifdef POSTGRES81
 extern DLLIMPORT char *DataDir;
 extern DLLIMPORT double cpu_operator_cost;
-#endif /* defined(POSTGRES81) && defined(WIN32) */
+#endif /* POSTGRES81 */
+#ifdef POSTGRES82
+extern DLLIMPORT bool pgstat_collect_tuplelevel;
+#endif /* POSTGRES82 */
+#endif /* WIN32 */
 
 #if defined(POSTGRES82) || defined(POSTGRES83)
 #include "access/reloptions.h"
@@ -127,10 +132,27 @@ PG_FUNCTION_INFO_V1(pgs2options);
 PG_FUNCTION_INFO_V1(pgs2indexcache);
 
 
+typedef struct {
+  FmgrInfo filterFn;
+} support_funcs;
+
+typedef struct {
+  support_funcs funcs;
+  double indtuples;
+  index_info *index;
+  Oid oid_type;
+} build_stat;
+
+typedef struct {
+  sen_records *records;
+  index_info *index;
+  sen_recordh *mark;
+} scan_stat;
+
+
 inline static void *scan_stat_open(sen_records *records, index_info *ii);
 inline static void scan_stat_close(void *opaque);
 void __attribute__ ((constructor)) pgs2_init(void);
-const char *getludiaoption(const char *option);
 Datum pgs2build0(Relation heap, Relation index, IndexInfo *indexInfo,
                  int flags, Oid oid_type);
 Datum pgs2insert0(Relation r, Datum *values, bool *isnull,
@@ -139,8 +161,7 @@ Datum pgs2bulkdelete0(Relation r, IndexBulkDeleteCallback callback,
                       void *callback_state, IndexBulkDeleteResult *stats,
                       Oid oid_type);
 Datum pgs2contain0(Datum vt, Datum vq, Oid oid_type);
-Datum get_tuple_Datum(ExprContext *econtext, HeapTuple tup, int attnum,
-                      ExprState *estate);
+unsigned get_pseudo_pagesize(sen_index *index);
 
 static index_info *index_cache;
 static index_info *last_used_cache;
@@ -149,7 +170,6 @@ static int curr_score = 0;
 static int last_nhits = 0;
 static int max_n_index_cache = MAX_N_INDEX_CACHE;
 
-
 static void
 pgs2_logger_func(int level, const char *time, const char *title,
                  const char *msg, const char *location, void *func_arg)
@@ -162,13 +182,50 @@ static sen_logger_info pgs2_sen_logger_info = {sen_log_warning,
                                                SEN_LOG_TIME | SEN_LOG_MESSAGE,
                                                pgs2_logger_func};
 
+
 void __attribute__ ((constructor)) pgs2_init(void) {
   sen_rc rc;
+  char option[16];
+  snprintf(option, 16, "%d", MAX_N_SORT_RESULT);
+  SetConfigOption("ludia.max_n_sort_result", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%s", ENABLE_SEQSCAN);
+  SetConfigOption("ludia.enable_seqscan", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%d", SEQSCAN_FLAGS);
+  SetConfigOption("ludia.seqscan_flags", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%d", SEN_INDEX_FLAGS);
+  SetConfigOption("ludia.sen_index_flags", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%d", MAX_N_INDEX_CACHE);
+  SetConfigOption("ludia.max_n_index_cache", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%d", INITIAL_N_SEGMENTS);
+  SetConfigOption("ludia.initial_n_segments", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%s", USEGENERICCOST);
+  SetConfigOption("ludia.usegenericcost", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
+  snprintf(option, 16, "%f", SEGPAGEFACTOR);
+  SetConfigOption("ludia.segpagefactor", option,
+                  PGC_USERSET, PGC_S_DEFAULT);
   if (sen_logger_info_set(&pgs2_sen_logger_info)) {
     elog(INFO, "pgsenna2: cannot logger_info_set.");
   } else {
     elog(DEBUG1, "pgsenna2: logger_info_set.");
   }
+#ifdef WIN32
+  {
+    char mecabrc[MAXPGPATH];
+    sprintf(mecabrc, "%s", my_exec_path);
+    sprintf(mecabrc + (strlen(mecabrc)-strlen("postgres.exe")), "mecabrc");
+    make_native_path(mecabrc);
+    if(!SetEnvironmentVariable("MECABRC", mecabrc)) {
+      elog(WARNING, "pgsenna2: cannot use mecab.");
+    }
+  }
+#endif
   rc = sen_init();
   if (rc != sen_success) {
     elog(ERROR, "pgsenna2: sen_init() failed %d", rc);
@@ -176,39 +233,6 @@ void __attribute__ ((constructor)) pgs2_init(void) {
 };
 
 
-const char*
-getludiaoption(const char *option)
-{
-  static int result = -1;
-
-  if (result == -1) {
-    const char *ccs = GetConfigOption("custom_variable_classes");
-    result = 0;
-    if (ccs != NULL) {
-      int dotPos = strlen("ludia");
-      const char *start = ccs;
-      for (;; ++ccs) {
-        int c = *ccs;
-        if (c == 0 || c == ',') {
-          if (dotPos == ccs - start && strncmp(start, "ludia", dotPos) == 0) {
-            result = 1;
-            break;
-          }
-          if (c == 0) {
-            break;
-          }
-          start = ccs + 1;
-        }
-      }
-    }
-  }
-  if (result) {
-    return GetConfigOption(option);
-  }
-  return NULL;
-}
-
-
 void
 sen_check_init(void)
 {
@@ -271,28 +295,6 @@ index_list_update(index_info *ii) {
 }
 
 
-inline static sen_encoding
-get_sen_encoding()
-{
-  sen_encoding encoding;
-
-  switch (GetDatabaseEncoding()) {
-  case PG_UTF8:
-    encoding = sen_enc_utf8;
-    break;
-  case PG_EUC_JP:
-    encoding = sen_enc_euc_jp;
-    break;
-  case PG_SJIS:
-    encoding = sen_enc_sjis;
-    break;
-  default:
-    encoding = sen_enc_default;
-  }
-  return encoding;
-}
-
-
 /* todo: SizeOfIptrData may be more suitable.  */
 #define KEY_SIZE (sizeof(ItemPointerData))
 
@@ -309,26 +311,26 @@ index_info_open(Relation r, int createp, int flags)
 
   sen_check_init();
   ii = last_used_cache;
-  for (i = max_n_index_cache - 1; i; i--) {
-    if(!ii) {
-      break;
-    } else {
-      if (ii->index &&
-          namespace == ii->namespace &&
-          relid == ii->relid &&
-          !strcmp(ii->name, name)) {
-        if (createp) { /* for REINDEX (and CLUSTER) */
-          break;
-        } else {
-          index_list_update(ii);
-          return ii;
-        }
+  /* get index_info from cache */
+  for (i = max_n_index_cache - 1; ii; i--) {
+    if (ii->index &&
+        namespace == ii->namespace &&
+        relid == ii->relid &&
+        !strcmp(ii->name, name)) {
+      if (createp) { /* for REINDEX (and CLUSTER) */
+        break;
       } else {
-        ii = ii->next;
+        index_list_update(ii);
+        return ii;
       }
+    } else {
+      if (!i) {
+        break;
+      }
+      ii = ii->next;
     }
   }
-  cand = ii;  /* cand is oldest or NULL */
+  cand = ii;  /* cand is least recently used or NULL */
   if (!cand) {
     for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
       if (!ii->using) {
@@ -338,9 +340,9 @@ index_info_open(Relation r, int createp, int flags)
     }
   }
   if (!cand) {
-    elog(ERROR, "pgsenna2: n of indices in use reached max(%d)", max_n_index_cache);
+    elog(ERROR, "pgsenna2: LRU algorithm failed (%d)", max_n_index_cache);
   }
-  index_list_update(cand);
+  index_list_update(cand); /* cand is least recently used or unused */
   sen_index_close(cand->index);
   if(cand->s && cand->s->opaque) {
     scan_stat_close(cand->s->opaque);
@@ -372,15 +374,14 @@ index_info_open(Relation r, int createp, int flags)
     }
     {
       RelFileNode *rels;
-      sen_index *sindex = NULL;
-      int flags_tmp, initial_n_segments_tmp;
 #ifdef POSTGRES83
       int num = smgrGetPendingDeletes(true, &rels, false);
 #else
       int num = smgrGetPendingDeletes(true, &rels);
 #endif /* POSTGRES83 */
-      char *pathname = (char *)palloc(sizeof(char) * MAXPGPATH);
+      char *pathname = (char *)PGS2_PALLOC(sizeof(char) * MAXPGPATH);
       for (i = 0; i < num; i++) {
+        sen_index *sindex = NULL;
         sprintf(pathname, "%s/%s/%d",
                 DataDir,
                 GetDatabasePath(rels[i].dbNode, rels[i].spcNode),
@@ -388,6 +389,7 @@ index_info_open(Relation r, int createp, int flags)
         sindex = sen_index_open(pathname);
         if (sindex) {
           sen_rc rc = sen_success;
+          int flags_tmp, initial_n_segments_tmp;
           sen_index_info(sindex, NULL, &flags_tmp, &initial_n_segments_tmp,
                          NULL, NULL, NULL, NULL, NULL, NULL, NULL);
           elog(DEBUG1, "pgsenna2: flags = %d->%d, initial_n_semengts = %d->%d",
@@ -432,7 +434,7 @@ index_info_close(index_info *ii)
 inline static void *
 scan_stat_open(sen_records *records, index_info *ii)
 {
-  scan_stat *ss = palloc(sizeof(scan_stat));
+  scan_stat *ss = PGS2_PALLOC(sizeof(scan_stat));
 
   ss->records = records;
   ss->index = ii;
@@ -484,141 +486,32 @@ init_support_funcs(support_funcs *funcs, Relation index)
                  CurrentMemoryContext);
 }
 
-
-/* while using array, section doesn't correspond with column-number. */
-static void
-update_index_with_Datum(sen_index *index, const void *key, unsigned int section,
-                        Datum oldv, Datum newv, Oid oid_type)
-{
-  sen_rc rc = sen_success;
-
-  if (oldv) {
-    if (oid_type == TEXTOID) {
-      sen_values *values = sen_values_open();
-      char *c = datum2cstr(oldv);
-      rc = sen_values_add(values, c, strlen(c), 0);
-      if (rc != sen_success) {
-        elog(ERROR, "pgsenna2: sen_values_add %d %d", rc, section);
-      }
-      rc = sen_index_update(index, key, section + 1, values, NULL);
-      if (rc != sen_success) {
-        elog(ERROR, "pgsenna2: sen_index_update failed %d,%d", rc, section);
-      }
-      rc = sen_values_close(values);
-      if (rc != sen_success) {
-        elog(ERROR, "pgsenna2: sen_values_close failed %d %d", rc, section);
-      }
-      pfree(c);
-    } else {
-      ArrayType* atype = DatumGetArrayTypeP(oldv);
-      char* v_ptr = ARR_DATA_PTR(atype);
-      int i;
-      for (i = 0; i < *ARR_DIMS(atype); i++) {
-        sen_values *values = sen_values_open();
-        Datum v = fetch_att(v_ptr, false, -1);
-        char *c = datum2cstr(v);
-        rc = sen_values_add(values, c, strlen(c), 0);
-        if (rc != sen_success) {
-          elog(ERROR, "pgsenna2: sen_values_add failed %d %d", rc, section);
-        }
-        rc = sen_index_update(index, key, i + 1, values, NULL);
-        if (rc != sen_success) {
-          elog(ERROR, "pgsenna2: sen_index_update failed %d,%d", rc, section);
-        }
-        rc = sen_values_close(values);
-        if (rc != sen_success) {
-          elog(ERROR, "pgsenna2: sen_values_close failed %d %d", rc, section);
-        }
-#if defined(POSTGRES81) || defined(POSTGRES82)
-        v_ptr = att_addlength(v_ptr, -1, PointerGetDatum(v_ptr));
-        v_ptr = (char *) att_align(v_ptr, 'i');
-#endif
-        /* SELECT typalign FROM pg_type WHERE typalign LIKE '%text%'; --'i' */
-        pfree(c);
-      }
-    }
-  }
-  if (newv) {
-    if (oid_type == TEXTOID) {
-      sen_values *values = sen_values_open();
-      char *c = datum2cstr(newv);
-      rc = sen_values_add(values, c, strlen(c), 0);
-      if (rc != sen_success) {
-        elog(ERROR, "pgsenna2: sen_values_add %d %d", rc, section);
-      }
-      rc = sen_index_update(index, key, section + 1, NULL, values);
-      if (rc != sen_success) {
-        elog(ERROR, "pgsenna2: sen_index_update failed %d,%d", rc, section);
-      }
-      rc = sen_values_close(values);
-      if (rc != sen_success) {
-        elog(ERROR, "pgsenna2: sen_values_close failed %d %d", rc, section);
-      }
-      pfree(c);
-    } else {
-      ArrayType* atype = DatumGetArrayTypeP(newv);
-      char* v_ptr = ARR_DATA_PTR(atype);
-      int i;
-      for (i = 0; i < *ARR_DIMS(atype); i++) {
-        sen_values *values = sen_values_open();
-        Datum v = fetch_att(v_ptr, false, -1);
-        char *c = datum2cstr(v);
-        rc = sen_values_add(values, c, strlen(c), 0);
-        if (rc != sen_success) {
-          elog(ERROR, "pgsenna2: sen_values_add failed %d %d", rc, section);
-        }
-        rc = sen_index_update(index, key, i + 1, NULL, values);
-        if (rc != sen_success) {
-          elog(ERROR, "pgsenna2: sen_index_update failed %d,%d", rc, section);
-        }
-        rc = sen_values_close(values);
-        if (rc != sen_success) {
-          elog(ERROR, "pgsenna2: sen_values_close failed %d %d", rc, section);
-        }
-#if defined(POSTGRES81) || defined(POSTGRES82)
-        v_ptr = att_addlength(v_ptr, -1, PointerGetDatum(v_ptr));
-        v_ptr = (char *) att_align(v_ptr, 'i');
-#endif
-        /* SELECT typalign FROM pg_type WHERE typalign LIKE '%text%'; --'i' */
-        pfree(c);
-      }
-    }
-  }
-}
-
-
 inline static void
 do_insert(Relation r, ItemPointer tid, index_info *ii,
           Datum *values, bool *isnull, support_funcs *funcs, Oid oid_type)
 {
-  char *c;
   sen_rc rc = sen_success;
-  Datum value;
   int i;
   int natts = RelationGetNumberOfAttributes(r);
 
   for (i = 0; i < natts; i++) {
     if (!isnull[i]) {
-      value = FunctionCall1(&funcs->filterFn, PointerGetDatum(values[i]));
-      c = datum2cstr(value);
+      Datum value = FunctionCall1(&funcs->filterFn, PointerGetDatum(values[i]));
+      char *c = datum2cstr(value);
       if (c && *c) {
-        LOCKTAG locktag;
         LockAcquireResult res = LOCKACQUIRE_OK;
+#ifdef POSTGRES81
+        LOCKTAG locktag;
         SET_LOCKTAG_RELATION(locktag, MyDatabaseId, ii->relid);
-#if defined(POSTGRES82) || defined(POSTGRES83)
-        res = LockAcquire(&locktag, ShareUpdateExclusiveLock, 0, 0);
-#else
         res = LockAcquire(DEFAULT_LOCKMETHOD, &locktag, r->rd_istemp,
                           ShareUpdateExclusiveLock, 0, 0);
 #endif
         if(res == LOCKACQUIRE_OK) {
-          update_index_with_Datum(ii->index, (void *)tid, i, 0, value, oid_type);
+          update_index_with_datum(ii, (void *)tid, i, 0, value, oid_type);
         } else {
           elog(ERROR, "pgsenna2: cannot LockAcquire while do_insert (%d)", res);
         }
-#if defined(POSTGRES82) || defined(POSTGRES83)
-        LockRelease(&locktag, ShareUpdateExclusiveLock, 0);
-#else
+#ifdef POSTGRES81
         LockRelease(DEFAULT_LOCKMETHOD, &locktag,
                     ShareUpdateExclusiveLock, 0);
 #endif
@@ -636,6 +529,11 @@ do_insert(Relation r, ItemPointer tid, index_info *ii,
       if (rc) {
         elog(ERROR, "pgsenna2: insert failed (%d)", rc);
       }
+    } else {
+      if (!(sen_sym_get(ii->index->keys, (void *)tid))) {
+        elog(ERROR, "pgsenna2: wrong tid while do_insert (%d)..",
+             sen_other_error);
+      }
     }
   }
 }
@@ -672,7 +570,7 @@ pgs2build0(Relation heap, Relation index, IndexInfo *indexInfo, int flags, Oid o
   reltuples = IndexBuildHeapScan(heap, index, indexInfo,
                                  buildCallback, (void *) &bs);
 #if defined(POSTGRES82) || defined(POSTGRES83)
-  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result = (IndexBuildResult *) PGS2_PALLOC(sizeof(IndexBuildResult));
   result->heap_tuples = reltuples;
   result->index_tuples = bs.indtuples;
   index_info_close(bs.index);
@@ -826,6 +724,7 @@ pgs2bulkdelete0(Relation r, IndexBulkDeleteCallback callback,
   IndexInfo* indexInfo = BuildIndexInfo(r);
   EState *estate = CreateExecutorState();
   TupleTableSlot *slot = MakeSingleTupleTableSlot(RelationGetDescr(hrel));
+  sen_rc rc;
 
   elog(DEBUG1, "pgsenna2: pgs2bulkdelete n=%d", nrecords);
   if (indexInfo->ii_Expressions != NIL &&
@@ -844,18 +743,32 @@ pgs2bulkdelete0(Relation r, IndexBulkDeleteCallback callback,
       if (heap_fetch(hrel, SnapshotAny, &htup, &hbuffer, false, NULL)) {
         (GetPerTupleExprContext(estate))->ecxt_scantuple = slot;
         update_index_with_tuple(indexInfo, ii, estate, &htup, NULL, oid_type);
+        ReleaseBuffer(hbuffer);
+        rc = sen_sym_del(ii->index->keys, &tid);
+        if (rc != sen_success) {
+          elog(ERROR, "pgsenna2: sen_sym_del failed while bulkdelete (%d)", rc);
+        }
+        tuples_removed += 1;
+      } else {
+#if defined(POSTGRES81) || defined(POSTGRES82)
+        elog(ERROR, "pgsenna2: heap tuple not found");
+#else
+        elog(WARNING, "pgsenna2: Ludia doesn't support VACUUM at PostgreSQL8.3, you should REINDEX ludia-index");
+        break;
+#endif
       }
-      ReleaseBuffer(hbuffer);
     }
   }
   FreeExecutorState(estate);
   ExecDropSingleTupleTableSlot(slot);
   heap_close(hrel, NoLock);
   if (stats == NULL) {
-    stats = (IndexBulkDeleteResult *) palloc0(sizeof(IndexBulkDeleteResult));
+    stats = (IndexBulkDeleteResult *) PGS2_PALLOC0(sizeof(IndexBulkDeleteResult));
   }
+  /*
   stats->num_pages = 1;
   stats->num_index_tuples = sen_sym_size(ii->index->keys);
+  */
   stats->tuples_removed = tuples_removed;
   PG_RETURN_POINTER(stats);
 }
@@ -902,11 +815,23 @@ _pgs2bulkdelete(PG_FUNCTION_ARGS)
 Datum
 pgs2vacuumcleanup(PG_FUNCTION_ARGS)
 {
-  /* IndexVacuumInfo *info = (IndexVacuumInfo *) PG_GETARG_POINTER(0); */
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  IndexVacuumInfo *info = (IndexVacuumInfo *) PG_GETARG_POINTER(0);
+#endif
   IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  index_info *ii = index_info_open(info->index, 0, 0);
 
-  /* under construction... */
-
+  if (stats == NULL) {
+    stats = (IndexBulkDeleteResult *) PGS2_PALLOC0(sizeof(IndexBulkDeleteResult));
+    stats->tuples_removed = 0;
+    stats->pages_removed = 0;
+    stats->pages_deleted = 0;
+    stats->pages_free = 0;
+  }
+  stats->num_pages = get_pseudo_pagesize(ii->index);
+  stats->num_index_tuples = sen_sym_size(ii->index->keys);
+#endif
   PG_RETURN_POINTER(stats);
 }
 
@@ -934,7 +859,18 @@ pgs2gettuple(PG_FUNCTION_ARGS)
   scan_stat *ss = s->opaque;
   sen_rc rc = sen_success;
   int res;
+  static CurrentId* currentid = NULL;
 
+  if (!currentid) {
+    currentid = start_currentid();
+  }
+  if (!comp_currentid(currentid)) {
+#ifdef POSTGRES83
+#else
+    pgstat_count_index_scan(&s->xs_pgstat_info);
+#endif
+    reset_currentid(currentid);
+  }
   if (dir != 1) {
     elog(WARNING, "pgsenna2: dir(%d) assigned in pgs2gettuple()", dir);
   }
@@ -971,7 +907,18 @@ pgs2getmulti(PG_FUNCTION_ARGS)
   int32 i;
   scan_stat *ss = s->opaque;
   bool more = true;
+  static CurrentId* currentid = NULL;
 
+  if (!currentid) {
+    currentid = start_currentid();
+  }
+  if (!comp_currentid(currentid)) {
+#ifdef POSTGRES83
+#else
+    pgstat_count_index_scan(&s->xs_pgstat_info);
+#endif
+    reset_currentid(currentid);
+  }
   if (!ss) {
     elog(ERROR, "pgsenna2: inconsistent scan, or max_n_index_cache is too small");
   }
@@ -999,12 +946,9 @@ pgs2rescan(PG_FUNCTION_ARGS)
   ScanKey    key = (ScanKey) PG_GETARG_POINTER(1);
   int i;
   sen_rc rc;
-  sen_query *q;
   sen_records *r;
   index_info *ii;
-  sen_encoding encoding;
   sen_sel_operator op;
-  const char *option;
   int max_n_sort_result = MAX_N_SORT_RESULT;
 
   elog(DEBUG1, "pgsenna2: pgs2rescan");
@@ -1015,10 +959,6 @@ pgs2rescan(PG_FUNCTION_ARGS)
 #endif
   scan_stat_close(s->opaque);
   ii = index_info_open(s->indexRelation, 0, 0);
-  rc = sen_index_info(ii->index, NULL, NULL, NULL, &encoding, NULL, NULL, NULL, NULL, NULL, NULL);
-  if (rc) {
-    elog(ERROR, "pgsenna2: index_info failed(%d)", rc);
-  }
   elog(DEBUG1, "pgsenna2: nok=%d ign=%d", s->numberOfKeys, s->ignore_killed_tuples);
   if (key && s->numberOfKeys > 0) {
     memmove(s->keyData, key, s->numberOfKeys * sizeof(ScanKeyData));
@@ -1039,7 +979,7 @@ pgs2rescan(PG_FUNCTION_ARGS)
     if(s->keyData[i].sk_argument != 0) {
       char *c;
       char *c_tmp = datum2cstr(s->keyData[i].sk_argument);
-      c = (char *) palloc (sizeof(char) * (strlen(c_tmp) + 64));
+      c = (char *) PGS2_PALLOC (sizeof(char) * (strlen(c_tmp) + 64));
       if (s->indexRelation->rd_index->indnatts == 1) {
         /* single column plain index or index on expressions */
         sprintf(c, "%s", c_tmp);
@@ -1053,40 +993,33 @@ pgs2rescan(PG_FUNCTION_ARGS)
         sprintf(c, "*W%d:5 %s", s->keyData[i].sk_attno, c_tmp);
       }
       elog(DEBUG1, "pgsenna2: query_text=(%s)", c);
-      if ((q = sen_query_open(c, strlen(c), sen_sel_or,
-                              PGS2_MAX_N_EXPRS, encoding))) {
-        if (sen_query_rest(q, NULL)) {
-          elog(WARNING, "pgsenna2: too many expressions (%d)",
-               sen_query_rest(q, NULL));
-        }
-        rc = sen_query_exec(ii->index, q, r, op);
-        if (rc && rc != sen_invalid_argument) {
-          elog(ERROR, "pgsenna2: sen_query_exec failed(%d)", rc);
-        }
-        rc = sen_query_close(q);
-        if (rc) {
-          elog(ERROR, "pgsenna2: sen_query_close failed(%d)", rc);
-        }
-      } else {
-        elog(ERROR, "pgsenna2: sen_query_open failed while pgs2rescan");
+      rc = sen_query_exec(ii->index, get_query(c), r, op);
+      if (rc && rc != sen_invalid_argument) {
+        elog(ERROR, "pgsenna2: sen_query_exec failed(%d)", rc);
       }
       pfree(c);
     }
   }
-  option = getludiaoption("ludia.max_n_sort_result");
-  if (option) {
-    max_n_sort_result = atoi(option);
-    if (max_n_sort_result < -1) {
-      elog(ERROR, "pgsenna2: value of max_n_sort_result is invalid: %d", max_n_sort_result);
+  {
+    const char *option = getludiaoption("ludia.max_n_sort_result");
+    if (option) {
+      max_n_sort_result = atoi(option);
+      if (max_n_sort_result < -1) {
+        elog(ERROR, "pgsenna2: value of max_n_sort_result is invalid: %d", max_n_sort_result);
+      }
+    } else {
+      elog(DEBUG1, "pgsenna2: value of max_n_sort_result = %d", max_n_sort_result);
     }
-  } else {
-    elog(DEBUG1, "pgsenna2: value of max_n_sort_result = %d", max_n_sort_result);
   }
   if (max_n_sort_result >= 0 && sen_records_nhits(r) > 0) {
     rc = sen_records_sort(r, max_n_sort_result, NULL);
     if (rc) {
       elog(ERROR, "pgsenna2: sort failed %d", rc);
     }
+    if (max_n_sort_result < sen_records_nhits(r)) {
+      elog(DEBUG1, "pgsenna2: max_n_sort_result is too small (%d<%d)",
+           max_n_sort_result ,sen_records_nhits(r));
+    }
   }
   s->opaque = scan_stat_open(r, ii);
   ii->s = s;
@@ -1146,12 +1079,12 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   IndexOptInfo *index = (IndexOptInfo *) PG_GETARG_POINTER(1);
   List *indexQuals = (List *) PG_GETARG_POINTER(2);
 #if defined(POSTGRES82) || defined(POSTGRES83)
-  //  RelOptInfo *outer_rel = (RelOptInfo *) PG_GETARG_POINTER(3);
+  RelOptInfo *outer_rel = (RelOptInfo *) PG_GETARG_POINTER(3);
   Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(4);
   Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(5);
   Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(6);
   double *indexCorrelation = (double *) PG_GETARG_POINTER(7);
-#else
+#elif defined(POSTGRES81)
   Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(3);
   Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(4);
   Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(5);
@@ -1163,11 +1096,58 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   double qual_op_cost;
   double qual_arg_cost;
   List *selectivityQuals;
-  /*
-    index_info *ii = index_info_open(index, 0, 0);
-    unsigned int nrecords = sen_sym_size(ii->index->keys);
-  */
+  bool usegenericcost;
+  double segpagefactor = SEGPAGEFACTOR;
 
+  {
+    const char *option;
+    char str[16];
+    option = getludiaoption("ludia.usegenericcost");
+    if (option) {
+      strncpy(str, option, 16);
+    } else {
+      strncpy(str, USEGENERICCOST, 16);
+    }
+    if (strcmp(str, "on") == 0) {
+      usegenericcost = true;
+    } else {
+      usegenericcost = false;
+    }
+    option = getludiaoption("ludia.segpagefactor");
+    if (option) {
+      segpagefactor = atof(option);
+    }
+  }
+
+  if (segpagefactor > 0.0) {
+#if defined(POSTGRES81)
+    Relation indrel = index_open(index->indexoid);
+#elif defined(POSTGRES82) || defined(POSTGRES83)
+    Relation indrel = index_open(index->indexoid, AccessShareLock);
+#endif
+    index_info *ii = index_info_open(indrel, 0, 0);
+#if defined(POSTGRES81)
+    index_close(indrel);
+#elif defined(POSTGRES82) || defined(POSTGRES83)
+    index_close(indrel, AccessShareLock);
+#endif
+    index->pages = (BlockNumber)(get_pseudo_pagesize(ii->index) * segpagefactor);
+  }
+
+  if (usegenericcost) {
+#if defined(POSTGRES82) || defined(POSTGRES83)
+    return DirectFunctionCall8(gincostestimate, PG_GETARG_DATUM(0),
+                               PG_GETARG_DATUM(1), PG_GETARG_DATUM(2),
+                               PG_GETARG_DATUM(3), PG_GETARG_DATUM(4),
+                               PG_GETARG_DATUM(5), PG_GETARG_DATUM(6),
+                               PG_GETARG_DATUM(7));
+#elif defined(POSTGRES81)
+    return DirectFunctionCall7(hashcostestimate, PG_GETARG_DATUM(0),
+                               PG_GETARG_DATUM(1), PG_GETARG_DATUM(2),
+                               PG_GETARG_DATUM(3), PG_GETARG_DATUM(4),
+                               PG_GETARG_DATUM(5), PG_GETARG_DATUM(6));
+#endif
+  }
   if (index->indpred != NIL)
     {
       List       *strippedQuals;
@@ -1182,7 +1162,7 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   *indexSelectivity = clauselist_selectivity(root, selectivityQuals,
                                              index->rel->relid,
                                              JOIN_INNER);
-  if (*indexSelectivity <= 1.0) { *indexSelectivity = 0.0; }
+  /* if (*indexSelectivity <= 1.0) { *indexSelectivity = 0.0; } */
   if (numIndexTuples <= 0.0)
     numIndexTuples = *indexSelectivity * index->rel->tuples;
   if (numIndexTuples > index->tuples)
@@ -1194,15 +1174,17 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   cost_qual_eval(&index_qual_cost, indexQuals);
 #endif
   qual_op_cost = cpu_operator_cost * list_length(indexQuals);
-  qual_arg_cost = index_qual_cost.startup +
-    index_qual_cost.per_tuple - qual_op_cost;
-  /* if (qual_arg_cost < nrecords) { qual_arg_cost = 0.0; } */
-  qual_arg_cost = 0.0;
+  qual_arg_cost = index_qual_cost.startup + index_qual_cost.per_tuple - qual_op_cost;
+  /* qual_arg_cost = 0.0; */
   *indexStartupCost = qual_arg_cost;
   *indexTotalCost += qual_arg_cost;
   *indexTotalCost += numIndexTuples * (cpu_index_tuple_cost + qual_op_cost);
 #if defined(POSTGRES82) || defined(POSTGRES83)
-  *indexTotalCost -= random_page_cost;
+  if (outer_rel != NULL && outer_rel->rows > 1) {
+    *indexTotalCost -= random_page_cost / ((double)outer_rel->rows);
+  } else {
+    *indexTotalCost -= random_page_cost;
+  }
 #endif
   *indexCorrelation = 1.0;
   elog(DEBUG1, "pgsenna2: cost=(%f,%f,%f)",
@@ -1216,85 +1198,101 @@ pgs2contain0(Datum vt, Datum vq, Oid oid_type)
 {
   char **ct, *cq;
   int res;
-  sen_query *sq;
   int i, ct_strlen_n = 1;
   unsigned int *ct_strlen;
-  int seqscan_flags = SEQSCAN_FLAGS;
+  static int seqscan_flags = SEQSCAN_FLAGS;
+  sen_rc rc;
+  static CurrentId* currentid = NULL;
 
-  {
-    const char *option;
-    char enable_seqscan[16];
-    elog(DEBUG1, "pgsenna2: pgs2contain");
-    option = getludiaoption("ludia.enable_seqscan");
-    if (option) {
-      strncpy(enable_seqscan, option, 16);
-    } else {
-      strncpy(enable_seqscan, ENABLE_SEQSCAN, 16);
+  if (!currentid) {
+    currentid = start_currentid();
+  }
+  if (!comp_currentid(currentid)) {
+    elog(DEBUG1, "pgsenna2: pgs2contain %d %d", currentid->xid, currentid->cid);
+    last_nhits = 0;
+    reset_currentid(currentid);
+    {
+      const char *option;
+      char enable_seqscan[16];
+      option = getludiaoption("ludia.enable_seqscan");
+      if (option) {
+        strncpy(enable_seqscan, option, 16);
+      } else {
+        strncpy(enable_seqscan, ENABLE_SEQSCAN, 16);
+      }
+      if (strcmp(enable_seqscan, "on")) {
+        ereport(ERROR, (errmsg("pgsenna2: sequencial scan disabled")));
+      }
     }
-    if (strcmp(enable_seqscan, "on")) {
-      ereport(ERROR, (errmsg("pgsenna2: sequencial scan disabled")));
+    {
+      const char *option;
+      option = getludiaoption("ludia.seqscan_flags");
+      if (option) {
+        seqscan_flags = atoi(option);
+      }
+      if (seqscan_flags < 0 || 1 < seqscan_flags) {
+        elog(ERROR, "pgsenna2: value of seqscan_flags is invalid: %d",
+             seqscan_flags);
+      }
     }
   }
-  {
-    const char *option;
-    option = getludiaoption("ludia.seqscan_flags");
-    if (option) {
-      seqscan_flags = atoi(option);
-    }
-    if (seqscan_flags < 0 || 1 < seqscan_flags) {
-      elog(ERROR, "pgsenna2: value of seqscan_flags is invalid: %d",
-           seqscan_flags);
-    }
+  if (!vt) {
+    PG_RETURN_BOOL(false);
   }
   if (oid_type == TEXTOID) {
-    ct = (char **) palloc(sizeof(char *));
+    ct = (char **) PGS2_PALLOC(sizeof(char *));
     *ct = datum2cstr(vt);
   } else {
-    ArrayType* vttype = DatumGetArrayTypeP(&vt);
-    Datum v;
+    ArrayType* vttype = DatumGetArrayTypeP(vt);
     char* v_ptr = ARR_DATA_PTR(vttype);
     ct_strlen_n = *ARR_DIMS(vttype);
-    ct = (char **) palloc(ct_strlen_n * sizeof(char *));
-    for(i = 0  ; i < ct_strlen_n; i++) {
-      v = fetch_att(v_ptr, false, -1);
-      ct[i] = datum2cstr(v);
-#if defined(POSTGRES81) || defined(POSTGRES82)
-      v_ptr = att_addlength(v_ptr, -1, PointerGetDatum(v_ptr));
-      v_ptr = (char *) att_align(v_ptr, 'i');
+    ct = (char **) PGS2_PALLOC(ct_strlen_n * sizeof(char *));
+#ifdef POSTGRES81
+    bits8 *bitmap = NULL;
+#else
+    bits8 *bitmap = ARR_NULLBITMAP(vttype);
 #endif
+    int bitmask = 1;
+    for(i = 0  ; i < ct_strlen_n; i++) {
+      if (!bitmap || (*bitmap & bitmask)) {
+        ct[i] = datum2cstr(fetch_att(v_ptr, false, -1));
+#if defined(POSTGRES81) || defined(POSTGRES82)
+        v_ptr = att_addlength(v_ptr, -1, PointerGetDatum(v_ptr));
+        v_ptr = (char *) att_align(v_ptr, 'i');
+#else
+        v_ptr = att_addlength_pointer(v_ptr, -1, PointerGetDatum(v_ptr));
+        v_ptr = (char *) att_align_nominal(v_ptr, 'i');
+#endif
+      } else {
+        ct[i] = PGS2_PALLOC(sizeof(char));
+        ct[i][0] = '\0';
+      }
+      if (bitmap) {
+        bitmask <<= 1;
+        if (bitmask == 0x100) {
+          bitmap++;
+          bitmask = 1;
+        }
+      }
     }
   }
   cq = datum2cstr(vq);
-  ct_strlen = (int *) palloc(ct_strlen_n * sizeof(int));
+  ct_strlen = (unsigned int *) PGS2_PALLOC(ct_strlen_n * sizeof(unsigned int));
   for (i = 0; i < ct_strlen_n; i++) {
     ct_strlen[i] = strlen(ct[i]);
   }
-  if ((sq = sen_query_open(cq, strlen(cq), sen_sel_or,
-                           PGS2_MAX_N_EXPRS, get_sen_encoding()))) {
-    sen_rc rc;
-    if (sen_query_rest(sq, NULL)) {
-      elog(WARNING, "pgsenna2: too many expressions (%d)",
-           sen_query_rest(sq, NULL));
-    }
-    rc = sen_query_scan(sq, (const char **)ct, ct_strlen, ct_strlen_n,
-                        seqscan_flags, &res, &curr_score);
-    /* if (rc != sen_success) { */
-    if (rc != sen_success && rc != sen_memory_exhausted) {
-      elog(ERROR, "pgsenna2: sen_query_scan failed (%d)", rc);
-    }
-    rc = sen_query_close(sq);
-    if (rc != sen_success) {
-      elog(ERROR, "pgsenna2: sen_query_close failed (%d)", rc);
-    }
-  } else {
-    elog(ERROR, "pgsenna2: sen_query_open failed while pgs2contain");
+  rc = sen_query_scan(get_query(cq), (const char **)ct, ct_strlen,
+                      ct_strlen_n, seqscan_flags, &res, &curr_score);
+  if (rc != sen_success) {
+    elog(ERROR, "pgsenna2: sen_query_scan failed (%d)", rc);
   }
   for (i = 0; i < ct_strlen_n; i++) {
     pfree(ct[i]);
   }
   pfree(ct);
   pfree(cq);
-  if (res) {
+  if (res && curr_score) {
+    last_nhits ++;
     PG_RETURN_BOOL(true);
   } else {
     PG_RETURN_BOOL(false);
@@ -1422,8 +1420,8 @@ pgs2indexcache(PG_FUNCTION_ARGS)
                       "that cannot accept type record")));
     attinmeta = TupleDescGetAttInMetadata(tupdesc);
     funcctx->attinmeta = attinmeta;
-    last_used_cache_array = (index_info **)palloc(max_n_index_cache *
-                                                  sizeof(index_info *));
+    last_used_cache_array = (index_info **) PGS2_PALLOC(max_n_index_cache *
+                                                        sizeof(index_info *));
     tmp_used_cache = last_used_cache;
     while(tmp_used_cache) {
       last_used_cache_array[n] = tmp_used_cache;
@@ -1449,9 +1447,9 @@ pgs2indexcache(PG_FUNCTION_ARGS)
     HeapTuple    tuple;
     Datum        result;
 
-    values = (char **) palloc(12 * sizeof(char *));
+    values = (char **) PGS2_PALLOC(12 * sizeof(char *));
     for (i = 0; i < 6; i++) {
-      values[i] = (char *) palloc(64 * sizeof(char));
+      values[i] = (char *) PGS2_PALLOC(64 * sizeof(char));
     }
     last_used_cache_array = (index_info **) funcctx->user_fctx;
     snprintf(values[0], 64, "%d", last_used_cache_array[call_cntr]->relid);
@@ -1473,104 +1471,32 @@ pgs2indexcache(PG_FUNCTION_ARGS)
 }
 
 
-Datum get_tuple_Datum(ExprContext *econtext,
-                     HeapTuple tup,
-                     int attnum,
-                     ExprState *estate)
+unsigned
+get_pseudo_pagesize(sen_index *index)
 {
-  Datum iDatum;
-  bool isNull;
+  unsigned file_size_keys;
+  unsigned file_size_lexicon;
+  unsigned long long inv_seg_size;
+  unsigned long long inv_chunk_size;
+  sen_rc rc;
 
-  ExecStoreTuple(tup, econtext->ecxt_scantuple, InvalidBuffer, false);
-  if (!estate)
-    iDatum = slot_getattr(econtext->ecxt_scantuple, attnum, &isNull);
-  else
-    iDatum = ExecEvalExprSwitchContext(estate, econtext, &isNull, NULL);
-  if (isNull)
-    {
-      iDatum = 0;
-    }
-  ExecClearTuple(econtext->ecxt_scantuple);
-  return iDatum;
-}
+  rc = sen_index_info(index,
+                      NULL, /* key_size */
+                      NULL, /* flags */
+                      NULL, /* initial_n_segments */
+                      NULL, /* encoding */
+                      NULL, /* nrecords_keys */
+                      &file_size_keys,
+                      NULL, /* nrecords_lexicon */
+                      &file_size_lexicon,
+                      &inv_seg_size,
+                      &inv_chunk_size);
 
+  if (rc != sen_success)
+    elog(ERROR, "pgsenna2: sen_index_info failed.");
 
-unsigned int update_index_with_tuple(IndexInfo* indexInfo,
-                                     index_info *ii,
-                                     EState *estate,
-                                     HeapTuple oldtup,
-                                     HeapTuple newtup,
-                                     Oid oid_type)
-{
-  ExprContext *econtext = GetPerTupleExprContext(estate);
-  int j;
-  unsigned int num_records = 0;
-  ListCell *indexpr_item = list_head(indexInfo->ii_ExpressionsState);
-
-  for (j = 0; j < indexInfo->ii_NumIndexAttrs; j++)
-    {
-      int keycol = indexInfo->ii_KeyAttrNumbers[j];
-      Datum oldDatum = 0;
-      Datum newDatum = 0;
-
-      if (keycol != 0)
-        { /* plain index */
-          if (oldtup)
-            oldDatum = get_tuple_Datum(econtext, oldtup, keycol, NULL);
-          if (newtup)
-            newDatum = get_tuple_Datum(econtext, newtup, keycol, NULL);
-        }
-      else
-        { /* index on expressions */
-          if (indexpr_item == NULL)
-            elog(ERROR, "pgsenna2: wrong number of index expressions");
-          if (oldtup)
-            oldDatum = get_tuple_Datum(econtext, oldtup, 0, (ExprState *)lfirst(indexpr_item));
-          if (newtup)
-            newDatum = get_tuple_Datum(econtext, newtup, 0, (ExprState *)lfirst(indexpr_item));
-          indexpr_item = lnext(indexpr_item);
-        }
-      if (oldDatum || newDatum)
-        {
-          LOCKTAG locktag;
-          LockAcquireResult res = LOCKACQUIRE_OK;
-
-          SET_LOCKTAG_RELATION(locktag, MyDatabaseId, ii->relid);
-#if defined(POSTGRES82) || defined(POSTGRES83)
-          res = LockAcquire(&locktag, ShareUpdateExclusiveLock, 0, 0);
-#else
-          /*
-            res = LockAcquire(DEFAULT_LOCKMETHOD, &locktag, r->rd_istemp,
-            ShareUpdateExclusiveLock, 0, 0);
-          */
-#endif
-          if(res == LOCKACQUIRE_OK)
-            {
-              if (oldDatum)
-                {
-                  update_index_with_Datum(ii->index, (&(oldtup->t_self)),
-                                          j, oldDatum, 0, oid_type);
-                }
-              if (newDatum)
-                {
-                  update_index_with_Datum(ii->index, (&(newtup->t_self)),
-                                          j, 0, newDatum, oid_type);
-                }
-            }
-          else
-            {
-              elog(ERROR, "pgsenna2: cannot LockAcquire while bulkdelete (%d)", res);
-            }
-          num_records += 1;
-#if defined(POSTGRES82) || defined(POSTGRES83)
-          LockRelease(&locktag, ShareUpdateExclusiveLock, 0);
-#else
-          /*
-            LockRelease(DEFAULT_LOCKMETHOD, &locktag,
-            ShareUpdateExclusiveLock, 0);
-          */
-#endif
-        }
-    }
-  return num_records;
+  return ((file_size_keys + SEN_SYM_SEGMENT_SIZE - 1) / SEN_SYM_SEGMENT_SIZE) +
+    ((file_size_lexicon + SEN_SYM_SEGMENT_SIZE - 1) / SEN_SYM_SEGMENT_SIZE) +
+    ((inv_seg_size + SEN_INV_SEGMENT_SIZE - 1) / SEN_INV_SEGMENT_SIZE) +
+    ((inv_chunk_size + SEN_INV_CHUNK_SIZE - 1) / SEN_INV_CHUNK_SIZE);
 }
